#!/usr/bin/python

################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#


#########################################################################
#
# valis -- ViewARCH lower information system.
#

LIB_DIR = None

import sys

if LIB_DIR:
  sys.path.insert(0, LIB_DIR)

import string
import cmd
import getopt
import tlatools
import vcache
import tc
import re
from tlatools import cfg

def is_control_file(f):

    ctl = '{arch}'
    if f[ : len(ctl) ] == ctl:
        return 1
    else:
        return 0

def get2list(a, cbv, r):

    cache_dir = cfg.general.tla_cachedir + '/' + a + '/' + cbv
    cf_f = tc.load_index(cache_dir, 'F')
    cf_d = tc.load_index(cache_dir, 'D')

    f_list = cf_f.check_out_to_list(r)
    d_list = cf_d.check_out_to_list(r)

    return [d_list, f_list]

def do_ls(cbv, item, a, r):

    if item == None or item == '':
        item = './'

    [d_list, f_list] = get2list(a, cbv, r)
    f_files = tlatools.get_files_by_directory_name(f_list, item, 0)
    d_files = tlatools.get_files_by_directory_name(d_list, item, 1)

    for i in range(len(d_files)):
        print d_files[i]
    for i in range(len(f_files)):
        print f_files[i]

def do_find(cbv, item, a, r):

    [d_list, f_list] = get2list(a, cbv, r)

    out_list = []
    for i in range(len(f_list)):
        m = re.match(r"^(\S+)\s", f_list[i])
        if m != None:
            f = m.group(1)
            if is_control_file(f) == 0:
                out_list.append(f)

    for i in range(len(d_list)):
        m = re.match(r"^(\S+)\s", d_list[i])
        if m != None:
            f = m.group(1)
            if is_control_file(f) == 0:
                out_list.append(f + '/')

    out_list.sort()
    print '\n'.join(out_list)

def get_chrono_list_generic(a, cbv, r, item, v_cmd):

    [d_list, f_list] = get2list(a, cbv, r)
    if len(item) >= 2 and item[:2] == './':
        item = item[2:]

    i_name = tlatools.find_assoc(f_list, item)
    if i_name == None:
        return []
    else:
        revs = tlatools.get_revisions(a, cbv)
        return vcache.get_inv_content(a, cbv, revs[r], i_name, v_cmd)

def do_cat(cbv, item, a, r):

    lines = get_chrono_list_generic(a, cbv, r, item, 0)
    for i in range(len(lines)):
        print lines[i],

def do_annotate(cbv, item, a, r):

    lines = get_chrono_list_generic(a, cbv, r, item, 1)
    for i in range(len(lines)):
        print "%3d: %s" % (lines[i][0], lines[i][1]),

def do_chrono_chart(cbv, item, a, r):

    lines = get_chrono_list_generic(a, cbv, r, item, 2)
    for i in range(len(lines)):
        s = lines[i][0]
        e = lines[i][1]
        if e == -1:
            s_e = "%d:" % s
        else:
            s_e = "(%d..%d):" % (s, e)

        print "%8.8s %s" % (s_e, lines[i][2]),

def do_inv_hist(cbv, item, a, r):

    iv_name = vcache.get_inv_name(a, cbv, r, item)

    ancestry_list = tlatools.get_ancestry(a, cbv)

    print "Inventory: %s" % iv_name
    print ""

    for i in range(len(ancestry_list)):

        abrowse_list = tlatools.get_abrowse(ancestry_list[i][0], ancestry_list[i][1])
        cf_f = vcache.load_inv_by_iname(ancestry_list[i][0], ancestry_list[i][1], iv_name)

        meta_info = cf_f.get_meta_info()

        for j in range(len(meta_info)):

            [commit_rev, f_name, f_type] = meta_info[ - (j+1) ]
            abr_idx = - (commit_rev + 1)

            print "  %s/%s--%s" % (ancestry_list[i][0], ancestry_list[i][1], abrowse_list[abr_idx][0])

            print "    Time:        %s" % abrowse_list[abr_idx][1]
            print "    Author:      %s" % abrowse_list[abr_idx][2]
            print "    Log Message: %s" % abrowse_list[abr_idx][3]
            print "    File Name:   %s" % f_name

            print ""

def usage():

    sys.stderr.write(\
'''
Name:

    valis - ViewARCH low level information system.

Synopsis:

    valis
    valis -h

        show this help

    valis [-A archive] { cat | ls | find | annotate | chrono-chart | inv-hist } \\
            { C--B--V | C--B--V--R } [DIR/FILE] 

Description:

    valis extracts various informations from ViewARCH's cache data, which is generated
    by gencache.py. The aim is to be able to use some feature from command line as well as CGI.

    Possible options are:

        -A ... select archive explicitly. default is `tla my-default-archive', as usual mannar.

    Possible commands are:

        cat ... cat a regular file. [DIR/FILE] must be FILE and can not be omitted.

        ls  ... directory listing. [DIR/FILE] must be DIR. if omitted, project root tree will
                be assumed.

        find .. list all dir/files recursively, just like find(1). at the moment, [DIR/FILE]
                must be omitted.

        annotate ...... print blame annotation chart. [DIR/FILE] must be FILE and can not be
                        omitted.

        chrono-chart .. print chrono chart. [DIR/FILE] must be FILE and can not be omitted.

        inv-hist ...... print inventory history.

    C--B--V--R means Category, Branch, Version, Revision, respectively.

''')


def _main():

    if len(sys.argv) == 1:
        usage()
        sys.exit(2)

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hA:", \
                 [
                   "help",
                   "archive"
                 ])
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    ###########################################################################
    # parse options
    ###########################################################################

    archive = None
    revision = None

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        if o in ("-A", "--archive"):
            archive = a

    if len(args) != 2 and len(args) != 3:
        usage()
        sys.exit()

    cmd = args[0]
    cbvr = args[1]
    m = re.match(r"^(.+--.+--.+)--(.+)$", cbvr)
    if m != None:
        cbv = m.group(1)
        revision = m.group(2)
    else:
        cbv = cbvr
        revision = None

    item = None
    if len(args) == 3:
        item = args[2]

    if archive == None:
        archive = tlatools.my_default_archive()

    if revision == None:
        revision = tlatools.get_latest_revision_no(archive, cbv)
    else:
        m = re.match(r"^(\d+)$", revision)
        if m == None:
            revision = tlatools.get_revision_no_from_name(archive, cbv, revision)
        else:
            revision = int(m.group(1))
            if revision > tlatools.get_latest_revision_no(archive, cbv):
                raise Exception

    try:
        eval('do_%s(cbv, item, archive, revision)' % cmd.replace('-', '_'))
    except NameError:
        print "Error: command '%s' not found." % cmd


if __name__ == '__main__': _main()

# arch-tag: 811b21c0-7086-472c-b9d4-f51ad9506273

