#!/usr/bin/python

################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#

################################################################
#
# vchart.py -- chrono chart handler.
#

import re
import sys
import tlatools
import os
import stat
import gzip
import string
try:
	import tempfile
	tempfile.mkstemp
except (ImportError, AttributeError):
	import vtempfile as tempfile


patch_file_RE = re.compile(r"^@@\s+\S+\s+\+(\d+).*@@")

def get_d(dd):
 
    m = re.match(r"^(\d+)$", dd)
    if m:
        return int(m.group(1))

    return None

class meta_event_info:

    def __init__(self):

        self.revl = []

    def read(self, revs):

        list = string.split(revs, ' ')
        for i in range(len(list)):
            self.revl.append(tlatools.get_dd(list[i]))

    def write(self):

        for i in range(len(self.revl)):
            crl = self.revl[i]
            if crl[0] == crl[1]:
                if i == 0:
                    commit_revision_list = str(crl[0])
                else:
                    commit_revision_list = commit_revision_list + ' ' + str(crl[0])
            else:
                if i == 0:
                    commit_revision_list = '%d..%d' % (int(crl[0]), int(crl[1]))
                else:
                    commit_revision_list = commit_revision_list + ' ' + '%d..%d' % (int(crl[0]), int(crl[1]))

        return commit_revision_list

    def append(self, rev):

        if len(self.revl) == 0:
            self.revl.append([rev, rev])
        else:
            last_rev = self.revl[-1][1]
            if rev == last_rev + 1:
                self.revl[-1][1] = last_rev + 1
            else:
                self.revl.append([rev, rev])

    def get_all_revs(self):

        revs = []
        for i in range(len(self.revl)):
            for j in range(self.revl[i][0], self.revl[i][1] + 1):
                revs.append(j)

        return revs

    def last(self):

        if len(self.revl) == 0:
            return -1
        else:
            return self.revl[-1][1]


class meta_state_info:
 
    def __init__(self, label, b_escape = 0):

        self.rlist = []
        self.label = label
        self.b_escape = b_escape

    def load(self, in_list, offset):

        datastr = in_list[offset][:-1]
        i = offset + 1
        while i < len(in_list) and in_list[i][0] in [' ', '\t', '\n']:
            datastr = datastr + in_list[i][:-1]
            i = i + 1

        ret_offset = i

        datalist = string.split(datastr)

        if len(datalist) == 0:
            return ret_offset

        self.label = datalist[0]

        i = 1
        while i < len(datalist):
            rr = datalist[i]
            if i + 1 < len(datalist):
                val = datalist[i+1]
            else:
                val = ''

            if self.b_escape == 1:
                val = tlatools.pika_unescape(val)

            d1 = get_d(rr)
            self.rlist.append([d1, val])

            i = i + 2

        return ret_offset


    def save(self):

        if self.label == '':
            return ''

        outstr = self.label
        if len(self.rlist) == 0:
            return outstr + '\n'

        for i in range(len(self.rlist)):
            [d1, val] = self.rlist[i]
            if self.b_escape == 1:
                val = tlatools.pika_escape(val)

            outstr = outstr + ' %d %s\n' % (d1, val)

        return outstr

    def putval(self, d, val):

        if len(self.rlist) == 0:
            self.rlist.append([d, val])
            return

        [d1, o_val] = self.rlist[-1]
        if d <= d1:
            raise Exception

        # if status is not changed, do nothing.
        if val == o_val:
            return

        self.rlist.append([d, val])

    def getval(self, d):

        if len(self.rlist) == 0:
            return None

        for i in range(len(self.rlist)):
            if d == self.rlist[i][0]:
                return self.rlist[i][1]
            if d < self.rlist[i][0]:
                if i == 0:
                    return None
                else:
                    return self.rlist[i-1][1]

        return self.rlist[-1][1]


def _wt(f, s):

     f.write(s)

class vchart:

    def __init__(self):

        self.lines = []
        self.commit_revision_list = meta_event_info()
        self.leol_status_list = meta_state_info('LEOL-STATUS-LIST:')
        self.file_name_history = meta_state_info('FILE-NAME-HISTORY:', 1)
        self.file_type_history = meta_state_info('FILE-TYPE-HISTORY:')
        self.file_permission_history = meta_state_info('FILE-PERMISSION-HISTORY:')
        self.loaded_file_name = ''

    def get_last_commit_revision(self):

        return self.commit_revision_list.last()

    def set_last_commit_revision(self, last_commit_revision):

        self.commit_revision_list.append(last_commit_revision)

    def get_loaded_file_name(self):

        return self.loaded_file_name

    def get_inventory_name(self, b_decode = 1):

        m = re.match(r".*/([a-zA-Z0-9=]+)(.gz)?$", self.loaded_file_name)
        if m:
            iv64 = m.group(1)
            if b_decode:
                return tlatools.iv64_2iv(iv64)
            else:
                return iv64
        else:
            return ''

    def get_meta_info(self):

        meta_info = []

        revs = self.commit_revision_list.get_all_revs()
        for i in range(len(revs)):
            rev = revs[i]
            meta_info.append([rev,
                              self.file_name_history.getval(rev),
                              self.file_type_history.getval(rev)])

        return meta_info

    def fload(self, f):

        ''' read header '''
        header_lines = []
        fl = f.readline()
        while fl:

            if fl[:-1] == '':
                break

            header_lines.append(fl)
            fl = f.readline()

        i = 0
        while i < len(header_lines):

            m = re.match(r"^COMMIT-REVISION-LIST: (.+)$", header_lines[i])
            if m:
                self.commit_revision_list.read(m.group(1))
                i = i + 1
                continue

            m = re.match(r"^LEOL-STATUS-LIST: (.+)$", header_lines[i])
            if m:
                i = self.leol_status_list.load(header_lines, i)
                continue

            m = re.match(r"^FILE-NAME-HISTORY: (.+)$", header_lines[i])
            if m:
                i = self.file_name_history.load(header_lines, i)
                continue

            m = re.match(r"^FILE-TYPE-HISTORY: (.+)$", header_lines[i])
            if m:
                i = self.file_type_history.load(header_lines, i)
                continue

            m = re.match(r"^FILE-PERMISSION-HISTORY: (.+)$", header_lines[i])
            if m:
                i = self.file_permission_history.load(header_lines, i)
                continue

            i = i + 1


        ''' read contents '''
        fl = f.readline()
        while fl:

            ''' for multiple vchart. '''
            if fl[:-1] == '':
                break

            m = re.match(r"^(\d+) (.*)$", fl)
            if m:
                s = int(m.group(1))
                l = m.group(2) + '\n'
                self.lines.append([s, -1, l])
            else:
                m = re.match(r"^\((\d+)\.\.(\d+)\) (.*)$", fl)
                if m:
                    s = int(m.group(1))
                    e = int(m.group(2))
                    l = m.group(3) + '\n'
                    self.lines.append([s, e, l])

            fl = f.readline()

    def load(self, chrono_file):

        self.loaded_file_name = chrono_file

        if os.access(chrono_file, os.F_OK):
            if chrono_file[-3:] == '.gz':
                f = gzip.open(chrono_file)
            else:
                f = open(chrono_file, 'r')
            self.fload(f)
            f.close()
            return 1
        else:
            return 0

    def fsave(self, f):

        _wt(f, 'COMMIT-REVISION-LIST: %s\n' % self.commit_revision_list.write())
        _wt(f, self.file_name_history.save())
        _wt(f, self.file_type_history.save())
        _wt(f, self.file_permission_history.save())
        _wt(f, self.leol_status_list.save())
       
        _wt(f, '\n')

        for i in range(len(self.lines)):

            [s, e, l] = self.lines[i]
            if e == -1:
                _wt(f, '%d %s' % (s, l))
            else:
                _wt(f, '(%d..%d) %s' % (s, e, l))

    def save(self, chrono_file = ''):

        if chrono_file == '':
            chrono_file = self.loaded_file_name

        fname = tlatools.make_tempfile_in_the_same_dir(chrono_file)

        if chrono_file[-3:] == '.gz':
            f = gzip.open(fname, 'wb', 9)
        else:
            f = open(fname, 'wb')

        self.fsave(f)
        if f != sys.stdout:
            f.close()
            os.chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
            tlatools.move_always(fname, chrono_file)

    def live_to_grand_offset(self, live_offset):

        active_line = 0

        for i in range(len(self.lines)):
            if self.lines[i][1] == -1:    # -1 means "the line is open."
                if active_line == live_offset:
                    return i
                else:
                    active_line = active_line + 1

        return len(self.lines)

    def get_next_grand_offset(self, grand_offset):

        for i in range(len(self.lines))[ grand_offset + 1: ]:
            if self.lines[i][1] == -1:    # -1 means "the line is open."
                return i

        return len(self.lines)

    def check_in_from_vanilla_list(self, list, label, seq):

        [f0, file_name] = tempfile.mkstemp()
        for i in range(len(list)):
            os.write(f0, list[i])
        os.close(f0)

        self.check_in_from_file(file_name, label, seq, 'T')

        os.remove(file_name)

    def patch_file(self, f_name, seq):

        f = open(f_name, 'r')

        live_offset = 0
        grand_offset = self.live_to_grand_offset(live_offset)

        df = f.readline()
        while df:
            if not df[:3] in ['---', '+++']:
                break

            df = f.readline()

        leol = self.leol_status_list.getval(seq - 1)

        while df:
            if df[:2] == '@@':

                # PATCH LOGIC.

                m = re.match(patch_file_RE, df)
                if m:
                    hunk_c = int(m.group(1))
                else:
                    raise Exception, 'bad patch format >>>%s<<<. exit' % df

                if hunk_c == 0:
                    live_offset = 0
                else:
                    live_offset = hunk_c - 1    # 0 - oriented line number

                grand_offset = self.live_to_grand_offset(live_offset)

                df = f.readline()

                prev_ch = None
                while df and not df[:2] == '@@':
                    if df[0] == '+':

                        self.lines.insert(grand_offset, [ seq, -1, df[1:] ])
                        live_offset = live_offset + 1
                        grand_offset = self.get_next_grand_offset(grand_offset)

                        prev_ch = '+'

                    elif df[0] == '-':
                        try:
                            self.lines[grand_offset][1] = seq - 1
                            grand_offset = self.get_next_grand_offset(grand_offset)
                            prev_ch = '-'
                        except Exception, val:
                            print 'grand_offset = %d' % grand_offset
                            print 'len(self.lines) = %d' % len(self.lines)
                            print '--------- self.lines --------------'
                            print ''.join(map(lambda x: x[2], self.lines[:10]))
                            print '--------- df --------------'
                            print ''.join(df)

                            print val

                            raise Exception

                    elif df[0] == ' ':
                        live_offset = live_offset + 1
                        grand_offset = self.get_next_grand_offset(grand_offset)
                        prev_ch = ' '

                    elif df[0] == '\\':

                        """
                        if prev_ch == ' ':
                            leols = "[-,-]"
                            leol = '0'
                        elif prev_ch == '+':
                            leols = "[+,-]"
                            leol = '0'
                        elif prev_ch == '-':
                            leols = "[-,+]"
                            leol = '1'
                        else:
                            leols = "[?,?]"
                            leol = '1'   # any way.

                        """

                        [leols, leol] = {
                            ' ':  [ "[-,-]", '0' ],
                            '+':  [ "[+,-]", '0' ],
                            '-':  [ "[-,+]", '1' ],
                            None: [ "[?,?]", '1' ]   # anyway.
                        }[prev_ch]

                        tlatools.tmsg("w leol %s (seq : %d, file : %s)" % \
                                (leols, seq, self.file_name_history.getval(seq)))
                        
                    df = f.readline()
            else:
                raise Exception, 'bad patch format. exit'

        f.close()

        self.leol_status_list.putval(seq, leol)

    '''
        append file to vchart.

        if append_file == None, empty file is assumed.
        label is used for file name history.
        seq is sequencial number.
        f_type is 'T', 'S', 'B', 'D', or None.
           if None, this function decides file type from append_file's
              information.
    '''
    def check_in_from_file(self, append_file, label, seq, f_type):

        # XXX: FILE-PERMISSION-HISTORY

        if f_type == None:
            if append_file == None:
                f_type = 'T'
            else:
                f_type = tlatools.f_type(append_file)

        ''' optimizing for initial check in '''
        if self.get_last_commit_revision() < 0:

            leol = '1'

            if f_type == 'T':

                if append_file != None:
                    f = open(append_file, 'r')
                    fl = f.readline()
                    while fl:
                        self.lines.append([ seq, -1, fl ])
                        fl = f.readline()
                    f.close()

                    if len(self.lines) > 0 and self.lines[-1][2][-1] != '\n':
                        self.lines[-1][2] = self.lines[-1][2] + '\n'
                        tlatools.tmsg('w leol = 0 (at initial import of "%s")' % label)
                        leol = '0'

            elif f_type == 'B':

                ''' not support yet. insert by base64 someday ? '''
                pass

            elif f_type == 'S':

                link_target = os.readlink(append_file) + '\n'
                self.lines.append([ seq, -1, link_target ])

            elif f_type == 'D':

                ''' can we append any type of useful info here ? '''
                pass

            else:

                raise Exception

            self.set_last_commit_revision(seq)
            self.file_name_history.putval(seq, label)
            self.file_type_history.putval(seq, f_type)
            self.leol_status_list.putval(seq, leol)

            return


        if append_file == None:
            append_file = '/dev/null'

        ''' ugly, but exception handling '''
        if f_type in ['B', 'D', 'S'] or \
           (f_type == 'T' and os.stat(append_file)[stat.ST_SIZE] == 0):
            for i in range(len(self.lines)):
                if self.lines[i][1] == -1:
                    self.lines[i][1] = seq - 1

            self.set_last_commit_revision(seq)
            self.file_name_history.putval(seq, label)
            self.file_type_history.putval(seq, f_type)
            self.leol_status_list.putval(seq, '1')

            return


        [f0, orig_file] = tempfile.mkstemp()
        os.close(f0)
        self.check_out_to_file(self.get_last_commit_revision(), orig_file)

        [f1, diff_out_file] = tempfile.mkstemp()
        os.close(f1)

        errmsg = 'diff -u %s %s > %s' % (orig_file, append_file, diff_out_file)
        tlatools.tsystem('diff -u %s %s > %s' % (orig_file, append_file, diff_out_file))

        os.remove(orig_file)

        ''' check seq '''
        if seq <= self.get_last_commit_revision():
            raise Exception, 'panic: invalid sequence no (%d <= %d) !!!' % (seq, self.get_last_commit_revision())

        ''' can't diff with prev rev. '''
        g = open(diff_out_file, 'r')
        gl = g.readline()
        if gl != None and len(gl) >= 6 and gl[:6] == 'Binary':
            self.set_last_commit_revision(seq)
            self.file_name_history.putval(seq, label)
            self.file_type_history.putval(seq, 'B')
            self.leol_status_list.putval(seq, 1)
            g.close()
            return
        g.close()

        ''' From now on, we can assume following.

               1. diff is succeeded.
               2. file type is 'T'
               3. file not empty
        '''
        self.patch_file(diff_out_file, seq)

        os.remove(diff_out_file)

        self.set_last_commit_revision(seq)
        self.file_name_history.putval(seq, label)
        self.file_type_history.putval(seq, 'T')

        ''' Because patch_file() sets leol status,
            Do not touch it here!
        '''
        # self.leol_status_list.putval(seq, 1)

    def check_out_to_list(self, seq_no, opt_format=0):

        # seq_no may be -1 (then return empty lines.)
        # opt_format -> (0 ... vanilla, 1 ... blame format, 2 ... chrono format, 3 ... full dump for debug)

        if opt_format == 3:

            ret_lines = []
            ret_lines.append('COMMIT-REVISION-LIST: %s\n' % self.commit_revision_list.write())
            ret_lines.append(self.file_name_history.save())
            ret_lines.append(self.file_type_history.save())
            ret_lines.append(self.file_permission_history.save())
            ret_lines.append(self.leol_status_list.save())
            ret_lines.append('\n')
            for i in range(len(self.lines)):
                ret_lines.append('%d, %d: %s' % (self.lines[i][0], self.lines[i][1], self.lines[i][2]))

            return ret_lines

        if opt_format == 2:          # chrono format
            return self.lines

        ret_lines = []
        for i in range(len(self.lines)):
            [s, e, l] = self.lines[i]

            if seq_no >= s and (e == -1 or seq_no <= e): # e == -1 means "the line is open".
                if opt_format == 0:  # vanilla
                    ret_lines.append(l)
                else:                # blame format.
                    ret_lines.append([s, l])

        if self.leol_status_list.getval(seq_no) == '0':
            if len(ret_lines) > 0:
                if len(ret_lines[-1]) > 0:
                    if opt_format == 0:
                        ret_lines[-1] = ret_lines[-1][:-1]
                    else:
                        ret_lines[-1][1] = ret_lines[-1][1][:-1]

        return ret_lines

    def check_out_to_file(self, seq_no, check_out_file):
        ret_lines = self.check_out_to_list(seq_no)
        f = open(check_out_file, 'w')
        for i in range(len(ret_lines)):
            f.write(ret_lines[i])
        f.close()

        # XXX: FILE-PERMISSION-HISTORY

def _main():

    print "nop."

if __name__ == '__main__': _main()

# arch-tag: 8cd4d3db-328e-4606-9c52-36e46237b71f
